program Arduino2RPIviaSerialP;

{

Arduino2RPIviaSerialP

	1.0		Nov 2014			Dan Hallock

This program communicates with an arduino via a serial (USB serial) port.

It sends 'commands' to the arduino to monitor an onboard switch and
motion detector.

If the switch is on and motion is detected, it will turn on an LED on the
arduino and it will also play a sound file on the PC.

It also sends a keep-alive packet to the arduino. The arduino uses this to
enable a different LED so you can verify that the arduino is receiving
data from the PC.

The commands that can be sent to the arduino are:
	0 - No command
	1 - Is Switch On?
	2 - Is motion detected?
	3 - Turn LED on
	4 - Turn LED off
	5 - Alive Packet

The arduino will respond by echoing back the command, and then including any
additional data after the command. This would be a T or F. For example, if
the PC sends '1' to the arduino, it would respond '1F' if the switch is off.

The commands are sent/received via the p2a_send and p2a_receive procedures.
These, in turn, call the p2a_i_ procedures which which 'packetize' the
command with checksum and the length of the string being transmitted. This
allows the arduino to verify the command was received error free.

The arduino uses the same process to send responses, so this program must
decode the packet and extract the command and response.

}

uses
	dateUtils,
    Classes,
//	MMSystem,
	strUtils,
    Synaser,
    Sysutils;

// ----------------------------------------------------------------------------
// global constants

const
	debug								= true;			// make true to see debug messages
	maxPacketLen						= 20;

// ----------------------------------------------------------------------------
// global variabls

var
    commPort                            : TBlockSerial;

// ----------------------------------------------------------------------------
// forward declarations

procedure initializeComm;
	forward;
function isMotionDetected				: boolean;
	forward;
function isSwitchOn						: boolean;
	forward;
procedure ledOff;
	forward;
procedure ledOn;
	forward;
function p2a_getStr(
	var	s								: string
	)									: boolean;
	forward;
procedure p2a_sendStr(
		s								: string
	);
	forward;
function p2a_i_computeChecksum(
		packet							: string
	)									: cardinal;
	forward;
function p2a_i_decodePacket(
		packet							: string;
	var	s								: string
	)									: boolean;
	forward;
procedure p2a_i_encodePacket(
		s								: string;
	var packet							: string
	);
	forward;
function p2a_i_getPacketLen(
		packet							: string
	)									: integer;
	forward;
function p2a_i_recvPacket(
	var	packet							: string
	)									: boolean;
	forward;
procedure p2a_i_xmitPacket(
		packet							: string
	);
	forward;
procedure playSound;
	forward;
procedure sendAliveMsg;
	forward;

// ----------------------------------------------------------------------------
// Setup parameters and initialize serial communications with the arduinoi

procedure initializeComm;

var
	baud								: string;
	n									: integer;
    port                        		: string;

begin

while (true) do begin
    write('Enter comm port to use [/dev/ttyACM0]: ');
    readln(port);
	if (length(port) = 0) then
		port := '/dev/ttyACM0';
	// everything is good
	break;
    end; // while

writeln;
writeln('Enter the Arduino''s baud rate such as:');
writeln('1200	2400	4800	9600');
writeln('14400	19200	28800	38400');
writeln('56000	57600');
writeln;

while true do begin
	write('Enter baud rate[9600]:           ');
	readln(baud);
	if (length(baud) = 0) then
		baud := '9600';
	if (not tryStrToInt(baud, n)) then
		continue;
	// everything is good
	break;
	end; // while

commPort:=TBlockSerial.Create;

try
    commPort.Connect(port);
    commPort.config(strToInt(baud), 8, 'N', SB1, false, False);
except
    writeln('Error opening serial communications with Arduino.');
	halt;
	end;

// when the serial port goes live, the arduino is rebooted. Give it a few
// seconds before we start accessing it.

if (debug) then
	writeln('Waiting for arduino to initialize...');
sleep(4000);

// make sure there is nothing left hanging around in the buffers
commPort.Purge;
commPort.Flush;

end; // initializeComm

// ----------------------------------------------------------------------------
// send command to arduino to ask if it has detected motion recently

function isMotionDetected					: boolean;

var
	s									: string			= '';

begin

if debug then
	writeln('@isMotionDetected');

p2a_sendStr('2');

s := '';
if not p2a_getStr(s) then begin
	writeln('p2a_sendStr returned false');
	halt;
	end;

if (s[1] <> '2') then begin
	writeln('isMotionDetected received bad response');
	halt;
	end;

// the 2nd byte returned is a T(true) or F indicating if motion has occurred.
if (uppercase(s[2]) = 'T') then
	isMotionDetected := true
else
	isMotionDetected := false;

if debug then
	writeln('@isMotionDetected: returns ', ord(isMotionDetected));

end; // isMotionDetected

// ----------------------------------------------------------------------------
// send command to arduino to determine if the switch is on or off.

function isSwitchOn						: boolean;

var
	s									: string			= '';

begin

if debug then
	writeln('@isSwitchOn');

p2a_sendStr('1');

s := '';
if not p2a_getStr(s) then begin
	writeln('p2a_sendStr returned false');
	halt;
	end;

if (s[1] <> '1') then begin
	writeln('isSwitchOn received bad response');
	halt;
	end;

if (uppercase(s[2]) = 'T') then
	isSwitchOn := true
else
	isSwitchOn := false;

if debug then
	writeln('@isSwitchOn: returns ', ord(isSwitchOn));

end; // isSwitchOn

// ----------------------------------------------------------------------------
// send command to arduino to turn LED off

procedure ledOff;

var
	s									: string			= '';

begin

if debug then
	writeln('@ledOff');

p2a_sendStr('4');

s := '';
if not p2a_getStr(s) then begin
	writeln('p2a_sendStr returned false');
	halt;
	end;

if (s[1] <> '4') then begin
	writeln('ledOff received bad response');
	halt;
	end;

end; // ledOff

// ----------------------------------------------------------------------------
// send command to arduino to turn LED on

procedure ledOn;

var
	s									: string			= '';

begin

if debug then
	writeln('@ledOn');

p2a_sendStr('3');

s := '';
if not p2a_getStr(s) then begin
	writeln('p2a_sendStr returned false');
	halt;
	end;

if (s[1] <> '3') then begin
	writeln('ledOn received bad response');
	halt;
	end;

end; // ledOn

// ----------------------------------------------------------------------------
// get 'unpacketized' string from arduino (e.g. the header has been stripped)

function p2a_getStr(
	var	s								: string
	)									: boolean;

var
	packet								: string			= '';

begin;

if debug then
	writeln('@p2a_getStr');

s := '';

// get the next packet from the arduino
if (not p2a_i_recvPacket(packet)) then begin
	// if nothing received, return null string
	if debug then
		writeln('  rcvpacket returned error to getstr');
	p2a_getStr := false;
	exit;
	end;

// decode the packet. if any error, return false
if (not p2a_i_decodePacket(packet, s)) then begin
	// if decoding problem return error code
	if debug then
		writeln('  decodePacket returned false to getstr');
   p2a_getStr := false;
   exit;
   end;

p2a_getStr := true;

end; // p2a_getStr

// ----------------------------------------------------------------------------
// send 'unpacketized' string to arduino. (the header will be added)

procedure p2a_sendStr(
	s									: string
	);

var
	packet								: string			= '';

begin

if debug then
	writeln('@p2a_sendStr: s=', s);

p2a_i_encodePacket(s, packet);
p2a_i_xmitPacket  (packet);

end; // p2a_sendStr

// ----------------------------------------------------------------------------
// compute checksum of packet by xoring all bytes except the first

function p2a_i_computeChecksum(
		packet							: string
	)									: cardinal;

var
	cksum								: cardinal;
	k									: integer;

begin

if debug then
	writeln('p2a_i_computeChecksum: ');

cksum := 0;

for k := 2 to p2a_i_getPacketLen(packet) do begin
	cksum := cksum xor ord(packet[k]);
	end; // for

// Clear high order bit as it might have trouble being transmitted.
// also turn any control characters into a normal character.
cksum := cksum and $7F;
if (cksum < $20) then
    cksum := cksum or $20;

p2a_i_computeChecksum := cksum;

if debug then
	writeln('  cksum=', cksum, '/', chr(cksum));

end; // p2a_i_computeChecksum

// ----------------------------------------------------------------------------
// Verify packet is valid, then copy the payload to the output string. If any
// error occurs, return false.

function p2a_i_decodePacket(
		packet							: string;
	var	s								: string
	)									: boolean;

var
	k									: integer;
	len									: integer;
	cksum								: cardinal;

begin

if debug then
	writeln('@p2a_i_decode4Packet: packet=', packet);

s := '';
// verify physical packet isn't too short
if (length(packet) < 4) then begin
	// packet must be at least 4 bytes to be valid
	if debug then
		writeln('  length packet < 4');
	p2a_i_decodePacket := false;
	exit;
	end;

// verify imbedded len is within acceptable range
len := p2a_i_getPacketLen(packet);
if ((len < 3) or (len > maxPacketLen)) then begin
	p2a_i_decodePacket := false;
	exit;
	end;

// verify cksum received matches cksum computed
cksum := p2a_i_computeChecksum(packet);
if (chr(cksum) <> packet[1]) then begin
	p2a_i_decodePacket := false;
	exit;
	end;

// transfer the packet payload to s
for k := 4 to len do
	s := s + packet[k];

p2a_i_decodePacket := true;

end; // p2a_i_decodePacket

// ----------------------------------------------------------------------------
// Given a string, create a valid packet with header that can be transmitted to
// arduino.

procedure p2a_i_encodePacket(
		s								: string;
	var packet							: string
	);

var
    cksum								: cardinal;
    i									: integer;
    len									: integer;

begin

if debug then
	writeln('@p2a_i_encodePacket: s=', s);

// Create the packet with dummy header
packet := '000' + s;

// put len of packet into bytes 2&3
len    := length(packet);
for i := 3 downto 2 do begin
	packet[i] := chr((len mod 10) + ord('0'));
	len 	  := len div 10;
	end; // for

// put checksum in byte 1
cksum 	  := p2a_i_computeChecksum(packet);
packet[1] := chr(cksum);

end; // p2a_i_encodePacket

// ----------------------------------------------------------------------------
// extract the header packet len from the header

function p2a_i_getPacketLen(
		packet							: string
	)									: integer;

var
	i									: integer;
	len									: integer;

begin

if debug then
	writeln('@p2a_i_getPacketLen');

len := 0;
for i := 2 to 3 do begin
	if (not (packet[i] in ['0'..'9'])) then begin
		p2a_i_getPacketLen := -1;
		exit;
		end;
	len := len * 10;
	len := ord(packet[i]) - ord('0');
	end; // for

p2a_i_getPacketLen := len;

if debug then
	writeln(' returning len as ', len);

end; // p2a_i_getPacketLen

// ----------------------------------------------------------------------------
// receive packet from arduino
// I had a surprising amount of trouble with this procedure. I spent several
// hours trying to get recvByte or recvString to work. They seem straight
// forward yet they would immediately return with nothing or some of the data.
// Never could I capture everything coming from the arduino. I still have no
// explanation why recvString would not work, but by using waitingData
// (bytesUsed did NOT work in place of waitingData) and recvByte I was able
// to get this code to work.

function p2a_i_recvPacket(
	var	packet							: string
	)									: boolean;

var
	c									: byte;
	s									: string			= '';
	t									: TDateTime;

begin

if debug then
	writeln('@p2a_i_recvPacket');

t := incSecond(now, 5);

// initialize packet, then read bytes until LF occurs
packet := '';
while (true) do begin
	if compareDateTime(now, t) > 0 then begin
		writeln('Timeout waiting for arduino response.');
		halt;
		end;
	try
	    if commPort.WaitingData = 0 then
    	    sleep(50);
   	    c := commPort.RecvByte(100);
   	    case chr(c) of
   		    #10: break;
   		    #13: begin end;
   		    else
      		    s := s + chr(c);
   		    end; // case
	except
		if debug then
			writeln('Serial error occurred in p2a_i_recvPacket: ', commPort.LastErrorDesc);
		p2a_i_recvPacket := false;
		exit;
		end;
	end; // while

if debug then
	writeln('  s=', s);

packet 				:= s;
p2a_i_recvPacket 	:= true;

end; // p2a_i_recvPacket

// ----------------------------------------------------------------------------
// transmit packet to arduino. Must include CR at end of string.

procedure p2a_i_xmitPacket(
		packet							: string
	);

begin

if debug then
	writeln('@p2a_i_xmitPacket: packet=', packet);

commPort.SendString(packet + #13);

end; // p2a_i_xmitPacket

// ----------------------------------------------------------------------------
// this procedure manages playing a sound. I don't want the sound going crazy
// so this procedure will not allow the sound to be repeated within
// <minutesBetweenSound> minutes.

procedure playSound;

const
	initialized							: boolean = false;
	nextTime							: tDateTime = 0;
	minutesBetweenSound					= 1;

begin

if debug then
	writeln('@playSound');

if not initialized then begin
	if debug then
		writeln('  initializing');
	initialized := true;
	nextTime := now;
	end;

if compareDateTime(now, nextTime) > 0 then begin
	if debug then
		writeln('  ok to play the sound');
//	sndPlaySound('c:/tmp/klaxon-short.wav', snd_Async or snd_NoDefault);
	nextTime := incMinute(now, minutesBetweenSound);
	end;

end; // playSound

// ----------------------------------------------------------------------------
// send a keepAlive to the arduino. This just lets it know we are still
// around and it will keep its active LED lit.

procedure sendAliveMsg;

var
	s									: string			= '';

begin

if debug then
	writeln('@sendAliveMsg');

p2a_sendStr('5');

s := '';
if not p2a_getStr(s) then begin
	writeln('p2a_getStr returned false');
	halt;
	end;

if (s[1] <> '5') then begin
	writeln('sendAliveMsg received bad response');
	halt;
	end;

end; // sendAliveMsg

// ----------------------------------------------------------------------------
// Outer Block

begin

initializeComm;

while (true) do begin

	if isSwitchOn then
		if isMotionDetected then begin
			ledOn;
			playSound;
			end
		else
			ledOff
	else begin
		ledOff;
        break;
        end;

	sendAliveMsg;

	sleep(1*1000);											// normally check once a second
	end; // while

commPort.free;

end.

